#!/usr/bin/env python
"""
usage:
python mirhmm.py
-a,miRNA in Fasta format
-b,mRNA in Fasta format
-c,miRNA_expression; opitional
-p,number of cpu cores used',default=2
-o,output directory to store the predicted binding profile result
	
"""

import pdb
import sys
import os
import subprocess
import math
import uuid
import multiprocessing
import copy
import time
import gc
import argparse

#=======================================================================	
# general functions

def check_seed(seedx,x):
	dN={'A':'T','C':'G','G':'CT','U':'AG','T':'AG','N':'N'}
	seedx=[dN[item] for item in seedx]
	rc_seedx=seedx[::-1]
	flag=1
	for i in range(len(x)):
		if x[i] not in rc_seedx[i]:
			flag=0
			break 
	return flag


def parse_mout(lf):
	# parse miranda out
	#f=open(A,'r')
	#lf=f.readlines()
	#f.close()
	lf=lf.split('\n')
	#pdb.set_trace()
	lf=[item for item in lf if item[0:2]=='>>']
	lf=[item.split() for item in lf]
	dS={}
	for i in lf:
		dS[int(i[1])]=float(i[4])
	return dS
	
	
def score2prob(S):
	S=int(S)
	minS=120
	maxS=140
	if S<minS:
		return 0
	elif S>maxS:
		return 1
	else:
		score=float(S-minS)/(maxS-minS)
	return score

def isFound(key,L):
	key=int(key)
	DA=24 #24 nt
	for i in L:
		i=i.split(',')
		i=[int(i[0]),int(i[1])]
		if i[0]<=key and key<=i[1]:
			return True
		elif (abs(key-i[0])<=DA) or (abs(key-i[1])<=DA):
			return True
	return False
		
def prob_unbound(nt):
	# nt : observed nucleotide 
	bg=[0.29,0.21,0.29,0.21]
	if nt.upper() in 'ACGT':
		return bg['ACGT'.index(nt)]
	else:
		return 0.25	
			
def prob_bound(bij):
	C=-15
	ss=bij.seed_site()
	es=bij.energy_site(C)
	bs=filterBS(ss+es)
	A=[]
	
	input_mir='>'+bij.mir.name+'\n'+bij.mir.seq
	MS=[]# mRNA sequence
	for nt in range(len(bij.mr.seq)):
		if isFound(nt,bs):
			span=24
			seq_nt=bij.mr.seq[max(0,nt-span):nt]
			if len(seq_nt)>0:
				input_m='>'+str(nt)+'\n'+seq_nt
				MS.append(input_m)
	MS='\n'.join(MS)
	
	#==================================================================================
	sessionID='mr_'+str(uuid.uuid4())
	f=open('%s.mir'%(sessionID),'w')
	f.write(input_mir)
	f.close()
	
	f=open('%s.m'%(sessionID),'w')
	f.write(MS)
	f.close()
	
	#--------------------------------------------------------------------------------
	
	pp=subprocess.Popen(["miRanda\miranda.exe",'%s.mir'%(sessionID),'%s.m'%(sessionID)],stdout=subprocess.PIPE, stdin=subprocess.PIPE,stderr=subprocess.PIPE)
	#pdb.set_trace()
	mout,err=pp.communicate()
	pp.wait()
	#pdb.set_trace()
	sm=parse_mout(mout)
	#pdb.set_trace()
	"""
	#pp.terminate()
	os.system('miRanda\\miranda.exe %s.mir %s.m -sc 0 >%s.mout'%(sessionID,sessionID,sessionID))
	sm=parse_mout('%s.mout'%(sessionID))
	"""
	
	Sij=[]
	for nt in range(len(bij.mr.seq)):
		if nt in sm:
			Sij.append(sm[nt])
		else:
			Sij.append(0)
	Sij=[min(1.0,bij.mir.conc*score2prob(item)) for item in Sij]
	os.remove('%s.mir'%(sessionID))
	os.remove('%s.m'%(sessionID))
	#os.remove('%s.mout'%(sessionID))
	return Sij
	
#----------------------------------------------------------------------	
def filterBS(A):
	for i in range(len(A)-1):
		for j in range(i+1,len(A)):
			if A[i]!=[] and A[j]!=[]:
				if overlap(A[i],A[j]):
					A[i]=mergeBS(A[i],A[j])
					A[j]=[]
	A=[item for item in A if item!=[]]
	return A
	
	
def overlap(a,b):
	a=a.split(',')
	b=b.split(',')
	flag=(int(a[1])-int(b[0]))*(int(a[0])-int(b[1]))
	if flag<=0:
		return True
	else:
		return False

def mergeBS(a,b):
	a=a.split(',')
	b=b.split(',')
	return str(min(a[0],b[0]))+','+str(max(a[1],b[1]))

#----------------------------------------------------------------------
					
	
def readFile(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf="".join(lf).split('>')[1:]
	lf=[item.split('\n') for item in lf]
	lf=[[item[0],''.join(item[1:])] for item in lf]
	return lf
	
def readInfo(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf=[item.strip().split() for item in lf]
	return lf
	
def readEM(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf=[item.split('\t') for item in lf]
	return lf	
	
def mpHelper(hmm,x):
	# helper function for multiprocesssing
	return(hmm.calculateEMI(x))
	
def mpCHelper(x):
	return(hmm.calculateAll(x))
	
def mpGHelper(x):
	return(hmm.calculateGamma(x))
		
def normConc(A):
	# normalize the concentration to range [0,2]
	dA={}	
	for i in A:
		dA[i[0]]=float(i[1])
	minA=min(dA.values())
	maxA=max(dA.values())
	avgA=(minA+maxA)*0.5
	dA={item:dA[item]/avgA for item in dA}
	return dA

def getVal(A,a,b,c):
	if c not in A[a][b]:
		return 0
	else:
		return A[a][b][c]
		
def showMem():
	mem=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
	mem=mem/1000.
	return mem
	
#==========================================================================
class miRNA:
	def __init__(self,mi,mi_seq,conc=1.0):
		self.name=mi
		self.seq=mi_seq
		self.conc=conc
	
class mRNA:
	def __init__(self,m,m_seq):
		self.name=m
		self.seq=m_seq

class binding:	
	def __init__(self,mir,m):
		self.mir=mir
		self.mr=m
		
	def seed_site(self):
		seed=self.mir.seq[1:7]
		ss=[]
		for i in range(len(self.mr.seq)-len(self.mir.seq)+1):
			x=self.mr.seq[i:i+len(seed)]
			flag=check_seed(seed,x)
			if flag!=0:
				ss.append(i+len(seed))
		out=[]
		x=24
		for i in ss:
			out.append(str(max(0,i-x))+","+str(i+1))
		return out
	
	def energy_site(self,C):
		input_seq=">"+self.mir.name+"\n"+self.mir.seq+"\n"+">"+self.mr.name+"\n"+self.mr.seq
		#pdb.set_trace()
		pp1=subprocess.Popen(['Vienna\\RNAduplex.exe','-e','10','-s'],stdout=subprocess.PIPE, stdin=subprocess.PIPE,stderr=subprocess.PIPE)
		out,err=pp1.communicate(input=input_seq)
		pp1.wait()
		lf=[item for item in out.split('\n') if len(item)>0]
		lf=[item.strip() for item in lf if item[0]!=">"]
		lf=[item.split() for item in lf]
		lf=[item for item in lf if float(item[-1][1:-1])<C]
		lf=[item[-2].split(',') for item in lf]
		lf=[str(int(item[0])-1)+','+str(int(item[1])-1) for item in lf]
		return lf
#==================================================================================================
class HMM:
	def __init__(self,HST,OB):
		self.OB=OB   # Observation sequences , list of mRNA class instances
		self.ST=ST # list of miRNA class instances	
		self.n=len(self.ST)+1 # of hidden states
		self.PI=self.initPI()   # Init State probability
		self.TM=self.initTM()   # Transition Matrix
		self.mCount=0 # mRNA counter will be used
		self.sr=[0 for state in range(self.n)] # gamma sum for each of state i
		self.ss=[[0 for state in range(self.n)] for state in range(self.n)] # sigma sum for each of state i,j
		self.sp=[0 for state in range(self.n)] # pi sum  for each state i
		
	# initialize PI, initial probability of binding states
	def initPI(self):
		A=[round(1.0/self.n,3) for states in range(self.n)]
		return A
	
	
	# initialize TM, transition matrix
	def initTM(self):
		tm=[]
		for i in range(self.n):
			tm_i=[round(1.0/self.n,3) for states in range(self.n)]
			tm.append(tm_i)
		return tm		 
	#=========================================================================================================		
	# calculate-EM------------------------------------------------------
	def calculateEMI(self,i):
		# i is the mRNA instance
		em_i=[]
		# unbound states
		em_u=[]
		for k in range(len(i.seq)):
			em_uk=prob_unbound(i.seq[k])
			em_u.append(em_uk)
		em_i.append(em_u)

		# miRNA bound states
		for j in self.ST:
			# j: miRNAs
			bij=binding(j,i)
			em_ij=prob_bound(bij)
			em_i.append(em_ij)
		return em_i
		
	def calculateEM(self):
		tc=0 # mRNA counter
		T=len(self.OB)
		self.EM=[]
		while (tc<T):
			pool=multiprocessing.Pool(processes=P)
			res=[]
			for tp in xrange(P):
				if tc<T:
					mtc=self.OB[tc]
					res.append(pool.apply_async(mpHelper,args=(hmm,mtc,)))
					tc+=1
				else:
					break 
			pool.close()
			pool.join()
			#pdb.set_trace()
			for rt in res:
				tmp=rt.get()
				tl=[]
				for j in range(len(tmp)):
					dij={k:tmp[j][k] for k in range(len(tmp[j])) if tmp[j][k]!=0}
					tl.append(dij)
				self.EM.append(tl)
			
			
	#forward-algorithm-------------------------------------------------
	def alpha(self,t,m,atp):
		# calculate the alpha for all n states at position t of mRNA m
		# atp: alpha for all n states at position t-1 of mRNA m
		if t==0:
			A=[self.PI[state]*getVal(self.EM,m,state,0) for state in range(self.n)]
			A=[item/sum(A) for item in A]
			return A
			
		else:
			A=[]
			for j in range(self.n):
				TP=0
				for i in range(self.n):
					TP+=atp[i]*self.TM[i][j]
				TP=TP*getVal(self.EM,m,j,t)
				#TP=TP*self.EM[m][j][t]
				A.append(TP)
			A=[float(item)/sum(A) for item in A]
			return A
	
	# backward-algorithm------------------------------------------------
	def beta(self,t,m,btp):
		# calculate the beta for all n states at position t of mRNA m
		# btp: beta for all n states at postion t+1 of mRNA m
		T=len(self.OB[m].seq)-1
		if t==T:
			B=[1 for state in range(self.n)]
			return B
		else:
			B=[]
			for i in range(self.n):
				TP=0
				for j in range(self.n):
					TP+=self.TM[i][j]*getVal(self.EM,m,j,t+1)*btp[j]
				B.append(TP)	
			B=[item/sum(B) for item in B]
			return B 

	#calculate gamma----------------------------------------------------------------
	def gamma(self,t,m,Am,Bm):
		# calculate the gamma for all n states at postion t of mRNA m
		GM=[]
		for i in range(self.n):
			rt=Am[t][i]*Bm[t][i]
			GM.append(rt)
		GM=[float(item)/sum(GM) for item in GM]
		return GM
	
	# calculate sigma
	def sigma(self,t,m,Am,Bm):
		# calculate the sigma from for all n states at postion t->t+1 of mRNA m
		SIG=[]	
		S=0
		for i in range(self.n):
			SIGI=[]
			for j in range(self.n):
				sigma_ij=(Am[t][i])*(self.TM[i][j])*(getVal(self.EM,m,j,t+1))*(Bm[t+1][j])
				S+=sigma_ij
				SIGI.append(sigma_ij)
			SIG.append(SIGI)
		
		for i in range(len(SIG)):
			SIG[i]=[float(item)/S for item in SIG[i]]
		return SIG
	
	def qx(self,m,Gm):		
		# calculate the maxmial linklihood	
		MLM=0			
		T=len(self.OB[m].seq)
		for t in range(T):
			rmt=Gm[t]
			qmt=max(rmt)	
			MLM+=math.log(qmt)
		return MLM
		
	#-------------------------------------------------------------------	
	def calculateAll(self,m):
		# calculate all for mRNA m
		# m : mRNA m index ID
		#output: ss,sr,qxm
		sr=[0 for state in range(self.n)] # gamma sum for each of state i
		ss=[[0 for state in range(self.n)] for state in range(self.n)] # sigma sum for each of state i,j
		sp=[0 for state in range(self.n)] # pi sum  for each state i
		
		T=len(self.OB[m].seq)
		# calculate alpha
		atp=[]
		Am=[]
		for t in xrange(T):
			tp=self.alpha(t,m,atp)
			Am.append(tp)
			atp=tp
						
		# calculate beata	
		btp=[]
		Bm=[]
		for t in xrange(T-1,-1,-1):
			tp=self.beta(t,m,btp)
			Bm.append(tp)
			btp=tp
		Bm=Bm[::-1]
		

		# calculate gamma
		Gm=[]
		for t in xrange(T):
			rtm=self.gamma(t,m,Am,Bm)
			Gm.append(rtm)
			if t<T-1:
				for i in range(self.n):
					sr[i]+=rtm[i]
			
		for i in range(self.n):
			sp[i]+=Gm[0][i]
		
			
		# calculate sigma
		for t in xrange(T-1):
			stm=self.sigma(t,m,Am,Bm)
			for i in range(self.n):
				for j in range(self.n):
					ss[i][j]+=stm[i][j]
		
		# calculate QX 
		QXM=self.qx(m,Gm)
		return [QXM,sp,sr,ss]
		
		
	def estimatePI(self):
		# estimate PI
		T=len(self.OB)
		for i in range(self.n):
			self.PI[i]=self.sp[i]/T
		self.sp=[0 for state in range(self.n)] # pi sum  for each state i	
		
	def estimateTM(self):
		# estimate TM
		for i in range(self.n):
			for j in range(self.n):
				self.TM[i][j]=self.ss[i][j]/self.sr[i]
		self.sr=[0 for state in range(self.n)] # gamma sum for each of state i
		self.ss=[[0 for state in range(self.n)] for state in range(self.n)] # sigma sum for each of state i,j
	
	def accumulateE(self,sp,sr,ss):
		self.sp=[self.sp[ind]+sp[ind] for ind in xrange(len(self.sp))]
		self.sr=[self.sr[ind]+sr[ind] for ind in xrange(len(self.sr))]
		for xx in xrange(len(self.ss)):
			self.ss[xx]=[self.ss[xx][ind]+ss[xx][ind] for ind in xrange(len(self.ss[xx]))]
		
	def BW(self):
		print('starting EM...')
		# baum-welch algorithm
		PML=0
		#---------------------------------------------------
		T=len(self.OB)
		tc=0 # mRNA counter
		QX=0
		while (tc<T):
			pool=multiprocessing.Pool(processes=P)
			res=[]
			for tp in xrange(P):
				if tc<T:
					res.append(pool.apply_async(mpCHelper,args=(tc,)))
					tc+=1
				else:
					break 
			pool.close()
			pool.join()
			for rt in res:
				[QXM,sp,sr,ss]=rt.get()
				self.accumulateE(sp,sr,ss)
				QX+=QXM
				
		#------------------------------------------------
		ML=QX
		cut=200 # cutoff for EM estimation
		ct=0
		while(abs(ML-PML)>cut):
			print(ML)
			# M-setp
			self.estimatePI()
			self.estimateTM()
			# E-step
			PML=ML	
			#---------------------------------------------------
			tc=0 # mRNA counter
			QX=0
			while (tc<T):
				pool=multiprocessing.Pool(processes=P)
				res=[]
				for tp in xrange(P):
					if tc<T:
						res.append(pool.apply_async(mpCHelper,args=(tc,)))
						tc+=1
					else:
						break 
				pool.close()
				pool.join()
				for rt in res:
					[QXM,sp,sr,ss]=rt.get()
					self.accumulateE(sp,sr,ss)
					QX+=QXM
					
		#------------------------------------------------
			ML=QX
			ct+=1
			print('EM loop %s...'%(ct))
				
	def calculateGamma(self,m):
		#calculate all for mRNA m
		#
		# m : mRNA m index ID
		T=len(self.OB[m].seq)
		
		# calculate alpha
		atp=[]
		Am=[]
		for t in xrange(T):
			tp=self.alpha(t,m,atp)
			Am.append(tp)
			atp=tp
		
		# calculate beata	
		btp=[]
		Bm=[]
		for t in xrange(T-1,-1,-1):
			tp=self.beta(t,m,btp)
			Bm.append(tp)
			btp=tp
		Bm=Bm[::-1]
		
		
		# calculate gamma
		Gm=[]
		for t in xrange(T):
			rtm=self.gamma(t,m,Am,Bm)
			Gm.append(rtm)

		return Gm
			
#=======================================================================

if __name__=='__main__':
	parser=argparse.ArgumentParser()
	parser.add_argument('-a',metavar='miRNA',help="miRNA in Fasta format",required=True)
	parser.add_argument('-b', metavar='mRNA',help='mRNA in Fasta format',required=True)
	parser.add_argument('-c',nargs='?', metavar='miRNA_expression',help='opitional, miRNA expression level')
	parser.add_argument('-p',metavar='processors',nargs='?', help='number of cpu cores used',default=2,type=int)
	parser.add_argument('-o',metavar='output',help="output directory to store the predicted binding profile result",required=True)
	
	Args=parser.parse_args()
	P=Args.p
	# 
	global C
	C=-15 #energy cut
	# concentration
	if Args.c:
		mirc=readInfo(Args.c)
		dmirc=normConc(mirc)
	# miRNAs
	mirs=readFile(Args.a)
	ST=[]
	for i in mirs:
		mir_i=i[0].split(' ')[0]
		if Args.c:
			conci=dmirc[mir_i]
			mi=miRNA(mir_i,i[1],conci)
		else:
			mi=miRNA(mir_i,i[1])
		ST.append(mi)
	# mRNAs
	#pdb.set_trace()
	OB=[]
	mrnas=readFile(Args.b)
	for i in mrnas:
		m=mRNA(i[0],i[1])
		OB.append(m)
	
	global hmm	
	hmm=HMM(ST,OB)
	#------------------
	pdb.set_trace()
	print('calculating EM...')
	hmm.calculateEM()
	
	mir_name=Args.a.split(os.sep)[-1]
	m_name=Args.b.split(os.sep)[-1]
	
	# write out EM
	g=open(mir_name+'_'+m_name+'.EM','a')
	for m in hmm.EM:
		ti=[]
		for i in m:
			t=[str(k)+':'+str(i[k])for k in i]
			t=','.join(t)
			ti.append(t)
		ti='\t'.join(ti)
		g.write(ti+'\n')
	g.close()
	
	#read in EMM
	"""
	EM=readEM(sys.argv[5])
	EMM=[]
	for i in EM:
		EMMi=[]
		for j in i:
			j='{'+j.strip()+'}'
			dij=eval(j)
			EMMi.append(dij)
		EMM.append(EMMi)
	m1=showMem()
	hmm.EM=EMM
	del EMM
	"""
	#------------------------
	gc.collect()
	OB=copy.deepcopy(hmm.OB)
	#=======================================================================
	# approximation in EM stage, if # of microRNAs is large, limit the # of 
	# mRNAs for EM training. 
	# 
	print('estimating the missing parameters...')

	SizeLimit=100000
	Tlimit=SizeLimit/len(mirs)
	hmm.OB=hmm.OB[0:Tlimit]
	hmm.BW() #baum-welch algorithm
	#=======================================================================
	hmm.OB=OB
	#-------------------------------------------------------------------
	#output
	TM_out=copy.deepcopy(hmm.TM)
	for i in xrange(len(TM_out)):
		TM_out[i]=[str(item) for item in TM_out[i]]
		TM_out[i]='\t'.join(TM_out[i])
	TM_out='\n'.join(TM_out)
	f=open(mir_name+'_'+m_name+'.TM','w')
	f.write(TM_out)
	f.close()
	os.system("""gnuplot\\gnuplot.exe -e "i='%s_%s.TM';o='%s_%s.png'" heatmap.gnu"""%(mir_name,m_name,mir_name,m_name))

	# outputdir
	output=Args.o
	if os.path.exists(output)==False:
		os.mkdir(output)
	#=======================================================================
	# output binding profile

	pool=multiprocessing.Pool(processes=P)
	result=[]

	tc=0
	T=len(hmm.OB)
	print('outputing results...')
	while (tc<T):
		pool=multiprocessing.Pool(processes=P)
		res=[]
		mL=[] # mRNA index
		for tp in xrange(P):
			if tc<T:
				res.append(pool.apply_async(mpGHelper,args=(tc,)))
				mL.append(tc)
				tc+=1
			else:
				break 
		pool.close()
		pool.join()
		for k in xrange(len(mL)):
				m=mL[k]
				bp=res[k].get()
				for j in range(len(bp)):
					bp[j]=[str(item) for item in bp[j]]
					bp[j]='\t'.join(bp[j])
				bp='\n'.join(bp)
				f=open(output+'/'+hmm.OB[m].name+'.bp','w')
				f.write(bp)
				f.close()
				os.system("""gnuplot\\gnuplot.exe -e "i='%s/%s.bp';o='%s/%s.png'" heatmap.gnu"""%(output,hmm.OB[m].name,output,hmm.OB[m].name))
		
	print('done!')
